import ftplib
import re
import pandas as pd
import collections
import requests
import warnings
import subprocess
import sys
import tqdm
import io, os
from cmapPy.pandasGEXpress.GCToo import GCToo
import cmapPy.pandasGEXpress.write_gct as wg
import xml.etree.ElementTree as ET

PROTOCOL = "http://"
BASE_URL = "ftp.ebi.ac.uk"
dir_path = '/pub/databases/metabolights/studies/public/'
data_type = "Metabolomics"
dataset_source = "Metabolights"
STUDY_FILE = "study_data"
ASSAY_FILE = "assay_metadata"
SAMPLE_FILE = "sample_metadata"
xml_link = 'ftp://ftp.ebi.ac.uk/pub/databases/metabolights/eb-eye/eb-eye_metabolights_studies.xml'


def get_all_files(mtbls):
    all_files = []
    ftp = ftplib.FTP()
    ftp.connect(BASE_URL)
    ftp.login()
    ftp.cwd(dir_path + mtbls)
    ftp.dir(all_files.append)
    return all_files

def get_txt(all_files):
    txt_files = [re.split('20[0-9]+', f)[-1] for f in all_files if f.strip().split(' ')[-1].endswith('.txt')]
    return txt_files

def get_tsv(all_files):
    tsv_files = [re.split('20[0-9]+', f)[-1] for f in all_files if f.strip().split(' ')[-1].endswith('.tsv')]
    return tsv_files

def get_sample_file(all_files):
    sample_files = [f for f in all_files if f.strip().lower().startswith('s_')]
    return sample_files

def get_assay_file(all_files):
    assay_file = [f for f in all_files if f.strip().lower().startswith('a_')]
    return assay_file

def read_tsv_file(url):
    s = requests.get(url).content
    tsv = pd.read_csv(io.StringIO(s.decode('utf-8')),sep='\t')
    colnames = [str(x).rstrip("\n").strip() for x in tsv.columns.tolist()]
    return colnames

def read_sample_file(url):
    s=requests.get(url).content
    c=pd.read_csv(io.StringIO(s.decode('utf-8')), sep='\t')
    samples = [str(y).rstrip("\n").strip() for y in c['Sample Name'].tolist()]
    return samples

def read_txt_file(url):
    s=requests.get(url).content
    c=pd.read_csv(io.StringIO(s.decode('utf-8')), sep='\t')
    samples = [str(y).rstrip("\n").strip() for y in c['Sample Name'].tolist()]
    return samples


def map_assay_sample_tsv_file(study, tsv_files, assay_files, sample_files):
    # mapping of assay file with tsv file
    tsv_mapped, assay_mapped, sample_mapped = [], [], []
    study_dataframe = pd.DataFrame(columns = [STUDY_FILE, ASSAY_FILE, SAMPLE_FILE])

    for tsv_file in tsv_files:
        tsv_url = PROTOCOL + BASE_URL + dir_path + study + "/" + tsv_file.lstrip()
        tsv_file_header = read_tsv_file(tsv_url)
        # print(tsv_file_header)

        for assay_file in assay_files:
            if tsv_file.split('_v2_maf.tsv')[0].split('m_')[-1] in assay_file:
                assay_url = PROTOCOL + BASE_URL + dir_path + study + "/" + assay_file.lstrip()
                assay_samples = read_txt_file(assay_url)
                check = all(item in tsv_file_header for item in assay_samples)
                if check == True:
                    tsv_mapped.append(tsv_url)
                    assay_mapped.append(assay_url)
                    # mapping of assay file with sample file
                    for sample_file in sample_files:
                        sample_file_url = PROTOCOL + BASE_URL + dir_path + study + "/" + sample_file.lstrip()
                        study_samples = read_sample_file(sample_file_url)
                        if collections.Counter(assay_samples) == collections.Counter(study_samples):
                            sample_mapped.append(sample_file)
                            if tsv_mapped and assay_mapped and sample_mapped:
                                file_mapping_dict = {}
                                file_mapping_dict[STUDY_FILE] = [tsv_file]
                                file_mapping_dict[ASSAY_FILE] = [assay_file]
                                file_mapping_dict[SAMPLE_FILE] = [sample_file]
                                df = pd.DataFrame.from_dict(file_mapping_dict)
                                study_dataframe = pd.concat([study_dataframe, df])
    
    return study_dataframe


def create_gct(parent_studies_df, mtbls):
    # Create GCT
    os.makedirs("GCTs", exist_ok = True)
    created_gcts = []

    print(f"Number of datasets for {mtbls} {parent_studies_df.shape[0]}")
    for index, row in tqdm.tqdm(parent_studies_df.iterrows()):
        study_name = mtbls
        study_file_link = BASE_URL + dir_path + mtbls + '/' + row[STUDY_FILE].lstrip()
        assay_meta_file_link = BASE_URL + dir_path + mtbls + '/' + row[ASSAY_FILE].lstrip()
        sample_meta_file_link = BASE_URL + dir_path + mtbls + '/' + row[SAMPLE_FILE].lstrip()
        
        try:
            files_to_download = [study_file_link, assay_meta_file_link, sample_meta_file_link]
            for file_link in files_to_download:
                file_link = file_link.replace(" ", "%20")
                try:
                    subprocess.check_output("wget " + file_link, shell=True, stderr=subprocess.STDOUT)
                except subprocess.CalledProcessError as e:
                    raise RuntimeError("command '{}' return with error {}): {}".format(e.cmd, e.returncode, e.output))

            # files to get data
            metabolite_file = study_file_link.split("/")[-1].replace("%20", " ")
            assay_file = assay_meta_file_link.split("/")[-1].replace("%20", " ")
            sample_file = sample_meta_file_link.split("/")[-1].replace("%20", " ")


            # read tsv file
            print("Reading file(s)")
            metabolite_data = pd.read_csv(metabolite_file, sep = '\t')
            metabolite_unique_cols = {}
            for metab_col in metabolite_data.columns:
                # remove spaces and new line if they are present in column names 
                modified_col = str(metab_col).replace(".", "_").rstrip("\n").strip()
                metabolite_unique_cols[metab_col] = modified_col
            metabolite_data = metabolite_data.rename(columns = metabolite_unique_cols)

            # renaming cols
            column_name_mapping = {"database_identifier" : "compoundId", "metabolite_identification" : "compound"}
            metabolite_data = metabolite_data.rename(columns = column_name_mapping)
            metabolite_data = metabolite_data[metabolite_data['compoundId'].notna()]

            # making row names unique
            row_descripter_freq = {}
            row_descripters = list(metabolite_data["compound"] + "_" + metabolite_data["compoundId"])
            # print(len(row_descripters))
            for index, row_name in enumerate(row_descripters):
                if row_name not in row_descripter_freq:
                    row_descripter_freq[row_name] = 0
                else:
                    row_descripter_freq[row_name] += 1
                    row_descripters[index] = str(row_descripters[index]) + "_" + str(row_descripter_freq[row_name])

            metabolite_data["row_des"] = row_descripters
            matrix_row_meta_df = metabolite_data.set_index('row_des')
            matrix_row_meta_df.index.name = None
            metabolite_data_columns = [str(x).rstrip('\n').strip() for x in matrix_row_meta_df.columns]


            # combining assay and sample metadata to get row metadata
            assay_col_metadata = pd.read_csv(assay_file ,delimiter = "\t")
            sample_metadata_df = pd.read_csv(sample_file ,delimiter = "\t")
            
            # renaming columns for assay and sample metadata df
            assay_new_columns = {}
            for column in list(assay_col_metadata.columns):
                if column != 'Sample Name':
                    assay_new_columns[column] = str("Assay_" + "_".join((column.split(" ")))).strip()

            sample_new_columns = {}
            for column in list(sample_metadata_df.columns):
                if column != 'Sample Name':
                    sample_new_columns[column] = str("Sample_" + "_".join((column.split(" ")))).strip()

            assay_col_metadata = assay_col_metadata.rename(columns = assay_new_columns)
            sample_metadata = sample_metadata_df.rename(columns = sample_new_columns)

            # merging sample metadata and assay metadata
            print("Merging sample metadata and assay metadata")
            sample_col_metadata = sample_metadata.merge(assay_col_metadata, on='Sample Name', how='left')

            sample_columns = {}
            for col in sample_col_metadata.columns:
                if "." in col:
                    sample_columns[col] = col.replace(".", " ")
            sample_col_metadata = sample_col_metadata.rename(columns = sample_columns)
            sample_col_metadata['Sample Name'] = sample_col_metadata['Sample Name'].apply(lambda x: x.rstrip('/n').strip())
            col_metadata = sample_col_metadata.set_index('Sample Name')
            col_metadata.index.name = None
            
            # get data matrix and row metadata df to prepare GCT
            print("Get data matrix and row metadata")
            samples = [str(col).rstrip('\n').strip() for col in sample_metadata_df["Sample Name"]]
            common_samples = [sample for sample in metabolite_data_columns if sample in samples]
            gct_matrix_data = matrix_row_meta_df.loc[:, common_samples]
            row_metadata_df = matrix_row_meta_df.loc[:, ~matrix_row_meta_df.columns.isin(samples)]

            # making gcts
            print("Making GCT")
            dataset_id = study_name + "_" + study_file_link.split("/")[-1].split("_v2_maf.tsv")[0]
            file_path = "GCTs" + "/" + dataset_id.replace("%20", "_").replace(" ", "_") + ".gct"
            final_gct_matrix_data = gct_matrix_data.sort_index(axis=1)
            col_metadata = col_metadata.sort_index(axis = 0)
            glyco_gctoo = GCToo(final_gct_matrix_data, row_metadata_df=row_metadata_df, col_metadata_df=col_metadata)
            wg.write(glyco_gctoo, file_path, data_null='NaN', metadata_null='', filler_null='')
            
            created_gcts.append(file_path)
        except Exception as e:
            print(e)
    return created_gcts

def delete_data(fformat):

    try:
        subprocess.check_output("rm -rf *."+fformat+"*", shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        raise RuntimeError("command '{}' return with error {}): {}".format(e.cmd, e.returncode, e.output))

    
def download_datasets(mtbls):
    file_list = get_all_files(mtbls)
    txt_files = get_txt(file_list)
    tsv_files = get_tsv(file_list)
    samples_files = get_sample_file(txt_files)
    assay_files = get_assay_file(txt_files)

    if tsv_files and assay_files and samples_files:
        parent_studies_df = map_assay_sample_tsv_file(mtbls, tsv_files, assay_files, samples_files)

        if not parent_studies_df.empty:
            # creating gct file
            gct_file_path = create_gct(parent_studies_df, mtbls)
            
            #Deleting all downloaded data
            delete_data('txt')
            delete_data('tsv')
            delete_data('xml')

            return gct_file_path
        else:
            warnings.warn('Failed to map files, exiting connector')
            sys.exit
    else:
        warnings.warn('Could not find files containing processed data, exiting connector')
        sys.exit


if __name__ == '__main__':
    gct_paths = download_datasets("MTBLS1051")
    print(gct_paths)